﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using System.Collections;
using System.Collections.ObjectModel;
using System.ComponentModel;

namespace HierarchyConverter.Wpf.Tests
{
    /// <summary>
    ///This is a test class for IHierarchyAwareTest and is intended
    ///to contain all IHierarchyAwareTest Unit Tests
    ///</summary>
    [TestClass]
    public class ReverseHierarchyConverterTests
    {
        [TestInitialize]
        public void TestSetup()
        {
            Converter = new ReverseHierarchyConverter<MyViewModel, int>(m => m.Id, m => m.ParentId, 0, -1);
            ItemsSource = MyViewModel.Gernerate();
            Converter.ItemsSource = ItemsSource;
        }

        public IList<MyViewModel> ItemsSource
        {
            get;
            set;
        }

        public ReverseHierarchyConverter<MyViewModel, int> Converter
        {
            get;
            set;
        }

        [TestMethod]
        [ExpectedException(typeof(InvalidOperationException))]
        public void ConverterShouldBeInitializedBeforeConversion()
        {
            var converter = new ReverseHierarchyConverter<MyViewModel, int>();
            var source = MyViewModel.Gernerate();
            converter.ItemsSource = source;
            converter.Convert(source.First(), typeof(IEnumerable), null, null);
        }

        [TestMethod]
        [ExpectedException(typeof(NotSupportedException))]
        public void ConvertsOnlyOneWay()
        {
            Converter.ConvertBack(null, null, null, null);
        }

        [TestMethod]
        public void IdentifiesPropertiesCorrectly()
        {
            Assert.AreEqual(0, Converter.RootParentIdValue);
            Assert.AreEqual(-1, Converter.UnassignedParentIdValue);

            Assert.AreEqual("Id", Converter.IdPropertyName);
            Assert.AreEqual("ParentId", Converter.ParentIdPropertyName);
        }

        [TestMethod]
        public void ChildrenColelctionsContainItemsWithTheCorrectParentId()
        {
            CheckHierarchyConsistency();
        }

        private void CheckHierarchyConsistency()
        {
            foreach (var item in ItemsSource)
            {
                var children = Converter.Convert(item, typeof(IEnumerable), null, null) as IList<MyViewModel>;
                Assert.IsNotNull(children);
                if (children.Any())
                {
                    Assert.IsTrue(children.All(child => child.ParentId == item.Id));
                }
            }
        }

        [TestMethod]
        public void TheRootItemsCollectionIsBuilt()
        {
            Assert.IsTrue(Converter.RootItems.Any());
            Assert.IsTrue(Converter.RootItems.All(item => item.ParentId == Converter.RootParentIdValue));
        }

        [TestMethod]
        public void TheUnassignedItemsCollectionIsBuilt()
        {
            Assert.IsTrue(Converter.UnassignedItems.Any());
            Assert.IsTrue(Converter.UnassignedItems.All(item => item.ParentId == Converter.UnassignedParentIdValue));
        }

        [TestMethod]
        public void NoPropertyChangedWhenItemsAreMovedInACollection()
        {
            var item = ItemsSource.First(i => i.Id == 4);
            var children = Converter.GetChildrenForItem(item) as ObservableCollection<MyViewModel>;
            children.Move(1, 3);
            Assert.AreEqual(0, ItemsSource.Sum(i => i.TimesPropertyChanged));
            this.CheckHierarchyConsistency();
        }

        [TestMethod]
        public void ItemIsUnassignedWhenChildIsRemoved()
        {
            var child = ItemsSource.First(i => i.Id == 4);
            var parent = ItemsSource.First(i => i.Id == child.ParentId);

            var children = Converter.GetChildrenForItem(parent) as IList<MyViewModel>;
            children.Remove(child);
            Assert.AreEqual(Converter.UnassignedParentIdValue, child.ParentId);
            this.CheckHierarchyConsistency();
        }

        [TestMethod]
        public void ItemIsAsignedWhenAChildIsInserted()
        {
            var parent = ItemsSource.First(i => i.Id == 4);
            var children = Converter.GetChildrenForItem(parent) as IList<MyViewModel>;
            var child = new MyViewModel() { Id = 100 };

            children.Add(child);
            Assert.AreEqual(parent.Id, child.ParentId);

            this.CheckHierarchyConsistency();
        }

        [TestMethod]
        public void SourceCollectionIsUpdatedWithNewlyInsertedItems()
        {
            var parent = ItemsSource.First(i => i.Id == 4);
            var children = Converter.GetChildrenForItem(parent) as IList<MyViewModel>;
            var child = new MyViewModel() { Id = 100 };

            children.Add(child);
            Assert.IsTrue(ItemsSource.Contains(child));
        }

        [TestMethod]
        public void SourceCollectionIsNotUpdatedWithExistignItems()
        {
            var parent = ItemsSource.First(i => i.Id == 4);
            var children = Converter.GetChildrenForItem(parent) as IList<MyViewModel>;
            var child = ItemsSource.First(i => i.Id == 9);

            children.Add(child);
            Assert.AreEqual(1, ItemsSource.Count(i => i.Id == 9));
        }

        [TestMethod]
        public void ParentIdChangeMovesTheItemToANewColelction()
        {
            var parent = ItemsSource.First(i => i.Id == 4);
            var children = Converter.GetChildrenForItem(parent) as IList<MyViewModel>;
            var child = ItemsSource.First(i => i.Id == 9);
            child.ParentId = parent.Id;

            Assert.IsTrue(children.Contains(child));
            Assert.IsFalse(Converter.UnassignedItems.Contains(child));
        }

        [TestMethod]
        public void RemovalsFromSourceCollectionAreRemoved()
        {
            var parent = ItemsSource.First(i => i.Id == 4);
            var children = Converter.GetChildrenForItem(parent) as IList<MyViewModel>;
            var child = children.First();
            ItemsSource.Remove(child);

            Assert.IsFalse(children.Contains(child));
            Assert.IsFalse(ItemsSource.Contains(child));
        }

        [TestMethod]
        public void AdditionsToSourceColelctionsAreAdded()
        {
            var parent = ItemsSource.First(i => i.Id == 4);
            var children = Converter.GetChildrenForItem(parent) as IList<MyViewModel>;
            var child = new MyViewModel() { Id = 100, ParentId = 4 };
            ItemsSource.Add(child);

            Assert.IsTrue(children.Contains(child));
            Assert.IsTrue(ItemsSource.Contains(child));
        }

        [TestMethod]
        public void ReplacementsInSourceCollectionAffectHierarchy()
        {
            var parent = ItemsSource.First(i => i.Id == 4);
            var children = Converter.GetChildrenForItem(parent) as IList<MyViewModel>;
            var newChild = new MyViewModel() { Id = 100, ParentId = 4 };
            var oldChild = children.Skip(1).First();
            var oldIndex = ItemsSource.IndexOf(oldChild);
            ItemsSource[oldIndex] = newChild;

            Assert.IsTrue(children.Contains(newChild));
            Assert.IsTrue(ItemsSource.Contains(newChild));
            Assert.IsFalse(children.Contains(oldChild));
            Assert.IsFalse(ItemsSource.Contains(oldChild));
        }

        [TestMethod]
        public void ClearingTheSourceCollectionAffectsHierarchy()
        {
            ItemsSource.Clear();
            Assert.IsFalse(Converter.RootItems.Any());
        }

        [TestMethod]
        public void ReplacementsInChildrenWorks()
        {
            var parent = ItemsSource.First(i => i.Id == 4);
            var children = Converter.GetChildrenForItem(parent) as IList<MyViewModel>;
            var newChild = ItemsSource.First(i => i.Id == 10);
            var oldChild = children.Skip(1).First();
            children[1] = newChild;

            Assert.IsTrue(children.Contains(newChild));
            Assert.IsTrue(Converter.UnassignedItems.Contains(oldChild));

            Assert.IsFalse(children.Contains(oldChild));
            Assert.IsTrue(ItemsSource.Contains(oldChild));
            Assert.IsFalse(Converter.UnassignedItems.Contains(newChild));
        }

        [TestMethod]
        public void ConvertingNonItemsYieldsNull()
        {
            Assert.IsNull(this.Converter.Convert("Not an item", typeof(IEnumerable), null, null));
        }

        [TestMethod]
        public void ClearingChildrenUnassignsThemAll()
        {
            var parent = ItemsSource.First(i => i.Id == 4);
            var children = Converter.GetChildrenForItem(parent) as IList<MyViewModel>;
            var savedChildren = children.ToList();
            children.Clear();
            
            Assert.IsTrue(ItemsSource.All(i => i.ParentId != 4));
            Assert.IsTrue(savedChildren.All(i => this.Converter.UnassignedItems.Contains(i)));
        }

        [TestMethod]
        public void ClearingTheUnassignedListChangesNothing()
        {
            var oldList = this.Converter.UnassignedItems.ToList();
            (this.Converter.UnassignedItems as IList<MyViewModel>).Clear();

            CollectionAssert.AreEqual(oldList, this.Converter.UnassignedItems as ICollection);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentException))]
        public void InitializingSelectorsMustBeMemberAccess()
        {
            var converter = new ReverseHierarchyConverter<MyViewModel, int>(m => 1, m => m.Title.Length, 0, -1);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentException))]
        public void InitializingSelectorsMustBePropertyAccess()
        {
            var converter = new ReverseHierarchyConverter<MyViewModel, int>(m => m.wrongId, m => m.ParentId, 0, -1);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentException))]
        public void InitializingSelectedPropertiesMustBeSettable()
        {
            var converter = new ReverseHierarchyConverter<MyViewModel, int>(m => m.Id, m => m.WrongParentId, 0, -1);
        }

        [TestMethod]
        public void InitializingTheConverterTwiceHasNoEffect()
        {
            var converter = new WrongConverter();
            converter.PublicInitialize(i => i.Id, i => i.ParentId, 1, -1);
            converter.PublicInitialize(i => i.ParentId, i => i.Id, 2, 3);

            Assert.AreEqual("Id", converter.IdPropertyName);
            Assert.AreEqual("ParentId", converter.ParentIdPropertyName);
            Assert.AreEqual(1, converter.RootParentIdValue);
            Assert.AreEqual(-1, converter.UnassignedParentIdValue);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentNullException))]
        public void TheIdSelectorCannotBeNull()
        {
            var converter = new ReverseHierarchyConverter<MyViewModel, int>(null, i => i.ParentId, 0, -1);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentNullException))]
        public void TheParentIdSelectorCannotBeNull()
        {
            var converter = new ReverseHierarchyConverter<MyViewModel, int>(i => i.Id, null, 0, -1);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentException))]
        public void TheUnassignedAndRootParentIdValuesCannotBeEqual()
        {
            var converter = new ReverseHierarchyConverter<MyViewModel, int>(i => i.Id, i => i.ParentId, 2, 2);
        }

        [TestMethod]
        public void TheItemsSourceCanBeChanged()
        {
            var oldSource = this.Converter.ItemsSource;
            var newSource = MyViewModel.Gernerate();

            this.Converter.ItemsSource = newSource;

            Assert.AreEqual(newSource, Converter.ItemsSource);
            Assert.IsTrue(Converter.RootItems.All(i => newSource.Contains(i)));
            Assert.IsTrue(Converter.UnassignedItems.All(i => newSource.Contains(i)));
        }

        [TestMethod]
        public void INotifyPropertyChangeWorks()
        {
            var converter = new WrongConverter();
            converter.PublicInitialize(i => i.Id, i => i.ParentId, 0, -1);

            var oldSource = converter.ItemsSource = this.Converter.ItemsSource;
            var newSource = MyViewModel.Gernerate();

            var expectedToRaise = new List<string>() { "ItemsSource", "RootItems", "UnassignedItems" };
            this.Converter.PropertyChanged += (sender, e) => expectedToRaise.Remove(e.PropertyName);

            this.Converter.ItemsSource = newSource;
            Assert.IsTrue(!expectedToRaise.Any());
        }
        
    }
}
